/*
 * File isocp.c - copy a file into an ISO9660 filesystem image
 *
 *
 * Copyright (C) 2006  R M Yorston
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * isocp - a program to copy a file into an ISO9660 filesystem image,
 *         replacing an existing file.
 *
 * Usage:
 *
 *    isocp imagefile file path
 *
 * where 'imagefile' is an ISO9660 image file
 *       'file' is the file to be written to the image
 *       'path' is the path to the file in the ISO9660 image.
 *
 * The target path can either specify the name of the file in the image
 * or the directory in which the file resides.  In the latter case the
 * name of the target file is assumed to be the same as that of the source
 * file.
 *
 * The file must already exist in the ISO image.  The replacement file can
 * always be smaller than the original.  It might be possible for the
 * replacement file to be larger than the original, but only if there is
 * some free space at the end of the last 2048 extent of the original file.
 *
 *
 * Changes:
 *
 * 2007-06-28  Add a couple of (off_t) casts.  Patch from Peter Joanes.
 *
 */
#define _FILE_OFFSET_BITS 64

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <malloc.h>

typedef unsigned char Uchar;
#include "iso9660.h"

int iso ;
unsigned char buffer[2048];

/* skip leading slashes */
const char *skip_slash(const char *s)
{
	for ( ; *s == '/' && *s != '\0'; ++s )
		;

	return s ;
}

/* strip trailing slashes */
void strip_slash(char *s)
{
	int len ;

	while ( (len=strlen(s)) && s[len-1] == '/' ) {
		s[len-1] = '\0' ;
	}
}

/* read a little-endian 32-bit quantity */
int isonum_733(char *p)
{
	return ((p[0] & 0xff)
		| ((p[1] & 0xff) << 8)
		| ((p[2] & 0xff) << 16)
		| ((p[3] & 0xff) << 24));
}

/* read a little-endian/big-endian 32-bit quantity */
void store_733(char *p, int val)
{
    p[0] = p[4] = val & 0xff ;
	p[1] = p[5] = (val >> 8) & 0xff ;
	p[2] = p[6] = (val >> 16) & 0xff ;
	p[3] = p[7] = (val >> 24) & 0xff ;
}


struct found_file {
	int extent ;		/* extent containing directory entry */
	int index ;			/* start of entry within extent */
} ;

/*
 * Attempt to find the specified path/file on the ISO image.  The return
 * value is a pointer to a found_file structure or NULL.  The global
 * variable 'buffer' contains a copy of the extent holding the directory
 * entry.
 */
struct found_file *find_file(const char *path, const char *file,
							int extent, int len)
{
	int i, slen ;
	struct iso_directory_record *idr ;
	char name_buf[256] ;
	char *sep ;

	if ( (sep=strchr(path, '/')) != NULL ) {
		/* we're looking for a directory */
		slen = sep - path ;
	}
	else {
		/* we're looking for a leaf */
		slen = strlen(path) ;
	}

	while ( len > 0 ) {
		lseek(iso, ((off_t)extent) << 11, SEEK_SET);
		read(iso, buffer, sizeof(buffer));

		len -= sizeof(buffer) ;
		++extent ;

		i = 0 ;
		while ( 1 ) {
			idr = (struct iso_directory_record *)&buffer[i];
			if ( idr->length[0] == 0 ) {
				break;
			}

			if (idr->name_len[0] == 1 && idr->name[0] == 0) {
				strcpy(name_buf, ".");
			}
			else if (idr->name_len[0] == 1 && idr->name[0] == 1) {
				strcpy(name_buf, "..");
			}
			else {
				strncpy(name_buf, idr->name, idr->name_len[0]);
				name_buf[idr->name_len[0]] = 0;
			}

			if ( strncmp(path, name_buf, slen) == 0 ) {
				/* we've found the path component we were looking for */
				if ( sep ) {
					/* found intermediate directory */
					if ( (idr->flags[0] & 2) ) {
						/* got directory, recurse down the path */
						return find_file(skip_slash(path+slen), file,
							isonum_733((unsigned char *)idr->extent),
	                		isonum_733((unsigned char *)idr->size));
					}
					else {
						/* expected directory, got file */
						return NULL ;
					}
				}
				else {
					/* found leaf */
					if ( (idr->flags[0] & 2) ) {
						/* leaf is a directory, look for named file */
						return find_file(file, file,
							isonum_733((unsigned char *)idr->extent),
							isonum_733((unsigned char *)idr->size));
					}
					else {
						/* leaf is a file */
						struct found_file *f ;

						f = (struct found_file *)
								malloc(sizeof(struct found_file));

						if ( f ) {
							f->extent = extent-1 ;
							f->index = i ;
						}

						return f ;
					}
				}
			}

			i += buffer[i];
			if (i > 2048 - offsetof(struct iso_directory_record, name[0])) {
				break;
			}
		}
	}

	return NULL ;
}

int main(int argc, char **argv)
{
	char *isoname ;
	char *filename ;
	char *pathname ;
	char *filepart ;
	char *s ;
	struct iso_primary_descriptor ipd ;
	struct iso_directory_record *idr ;
	struct found_file *f ;
	int fd ;
	struct stat buf ;
	off_t iso_size, iso_max_size ;
	int iso_extents ;
	int extent ;
	int count ;
	ssize_t bytes, i ;
	char filebuf[2048] ;
	off_t offset ;

	if ( argc != 4 ) {
		fprintf(stderr, "usage: %s iso file path\n", argv[0]) ;
		return 1 ;
	}

	isoname = argv[1] ;
	filename = argv[2] ;
	pathname = argv[3] ;

	/* get file part of filename */
	filepart = (s=strrchr(filename, '/')) != NULL ? s+1 : filename ;

	/* open source file and get its size */
	if ( (fd=open(filename, O_RDONLY)) < 0 ) {
		fprintf(stderr, "%s: failed to open file %s\n", argv[0], filename) ;
		return 1 ;
	}

	if ( fstat(fd, &buf) < 0 ) {
		fprintf(stderr, "%s: failed to stat file %s\n", argv[0], filename) ;
		return 1 ;
	}

	/* open ISO image */
	if ( (iso=open(isoname, O_RDWR)) < 0 ) {
		fprintf(stderr, "%s: failed to open ISO image %s\n", argv[0], isoname) ;
		return 1 ;
	}

	lseek(iso, ((off_t)16)<<11, SEEK_SET) ;
	read(iso, &ipd, sizeof(ipd)) ;

	idr = (struct iso_directory_record *)ipd.root_directory_record;

	if ((ipd.type[0] != ISO_VD_PRIMARY) ||
			(strncmp(ipd.id, ISO_STANDARD_ID, sizeof (ipd.id)) != 0) ||
			(ipd.version[0] != 1)) {
		fprintf(stderr, "%s: CD-ROM is not in ISO 9660 format\n", argv[0]);
		return 1 ;
	}

	/* try to find specified file in ISO image */
	strip_slash(pathname) ;
	f = find_file(skip_slash(pathname), filepart,
					isonum_733((unsigned char *)idr->extent),
	                isonum_733((unsigned char *)idr->size));

	if ( f == NULL ) {
		fprintf(stderr, "%s: failed to find file in ISO\n", argv[0]) ;
		return 1 ;
	}

	idr = (struct iso_directory_record *)&buffer[f->index] ;

	/* check that file will fit in ISO */
	iso_size = isonum_733((unsigned char *)idr->size) ;
	iso_extents = iso_size / 2048 ;
	if ( iso_size % 2048 ) {
		++iso_extents ;
	}
	iso_max_size = iso_extents * 2048 ;

	if ( buf.st_size > iso_max_size ) {
		fprintf(stderr, "%s: file is too big\n", argv[0]) ;
		return 1 ;
	}

	/* seek to file location in ISO */
	extent = isonum_733((unsigned char *)idr->extent) ;
	offset = (off_t)extent * 2048 ;
	if ( lseek(iso, offset, SEEK_SET) == (off_t)-1 ) {
		fprintf(stderr, "%s: failed to seek to extent\n", argv[0]) ;
		return 1 ;
	}

	/* write file to ISO */
	count = 0 ;
	while ( (bytes=read(fd, filebuf, 2048)) > 0 ) {
		/* pad buffer with nulls */
		for ( i=bytes; i<2048; ++i ) {
			filebuf[i] = '\0' ;
		}

		write(iso, filebuf, 2048) ;
		++count ;
	}

	/* fill buffer with nulls */
	for ( i=0; i<2048; ++i ) {
		filebuf[i] = '\0' ;
	}

	for ( ; count<iso_extents; ++count) {
		write(iso, filebuf, 2048) ;
	}

	close(fd) ;

	/* modify file size in directory entry */
	if ( buf.st_size != iso_size ) {
		store_733(idr->size, (int)buf.st_size) ;

		/* seek to directory location in ISO */
		offset = (off_t)f->extent * 2048 ;
		if ( lseek(iso, offset, SEEK_SET) == (off_t)-1 ) {
			fprintf(stderr, "%s: failed to seek to extent\n", argv[0]) ;
			return 1 ;
		}

		write(iso, buffer, 2048) ;
	}

	close(iso) ;

	return 0 ;
}
